//
// ServiceHostBase.cs
//
// Author:
//	Atsushi Enomoto <atsushi@ximian.com>
//
// Copyright (C) 2005-2006 Novell, Inc.  http://www.novell.com
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.ServiceModel.Description;
using System.ServiceModel.Dispatcher;
using System.ServiceModel.Security;
using System.Reflection;

namespace System.ServiceModel
{
	public abstract class ServiceHostBase
		: CommunicationObject, IExtensibleObject<ServiceHostBase>, IDisposable
	{
		ServiceCredentials credentials;
		ServiceDescription description;
		UriSchemeKeyedCollection base_addresses;
		TimeSpan open_timeout, close_timeout, instance_idle_timeout;
		ServiceThrottle throttle;
		List<InstanceContext> contexts;
		ReadOnlyCollection<InstanceContext> exposed_contexts;
		ChannelDispatcherCollection channel_dispatchers;
		IDictionary<string,ContractDescription> contracts;
		int flow_limit = int.MaxValue;
		IExtensionCollection<ServiceHostBase> extensions;

		protected ServiceHostBase ()
		{
			open_timeout = DefaultOpenTimeout;
			close_timeout = DefaultCloseTimeout;

			credentials = new ServiceCredentials ();
			throttle = new ServiceThrottle ();
			contexts = new List<InstanceContext> ();
			exposed_contexts = new ReadOnlyCollection<InstanceContext> (contexts);
			channel_dispatchers = new ChannelDispatcherCollection (this);
		}

		public event EventHandler<UnknownMessageReceivedEventArgs>
			UnknownMessageReceived;

		public ReadOnlyCollection<Uri> BaseAddresses {
			get { return new ReadOnlyCollection<Uri> (base_addresses.InternalItems); }
		}

		internal Uri CreateUri (string sheme, Uri relatieUri) {
			Uri baseUri = base_addresses.Contains (sheme) ? base_addresses [sheme] : null;

			if (relatieUri == null)
				return baseUri;
			if (relatieUri.IsAbsoluteUri)
				return relatieUri;
			if (baseUri == null)
				return null;
			return new Uri (baseUri, relatieUri);
		}

		[MonoTODO]
		public ChannelDispatcherCollection ChannelDispatchers {
			get { return channel_dispatchers; }
		}

		public ServiceAuthorizationBehavior Authorization {
			get;
			private set;
		}

		[MonoTODO]
		public ServiceCredentials Credentials {
			get { return credentials; }
		}

		public ServiceDescription Description {
			get { return description; }
		}

		protected IDictionary<string,ContractDescription> ImplementedContracts {
			get { return contracts; }
		}

		[MonoTODO]
		public IExtensionCollection<ServiceHostBase> Extensions {
			get {
				if (extensions == null)
					extensions = new ExtensionCollection<ServiceHostBase> (this);
				return extensions;
			}
		}

		protected internal override TimeSpan DefaultCloseTimeout {
			get { return DefaultCommunicationTimeouts.Instance.CloseTimeout; }
		}

		protected internal override TimeSpan DefaultOpenTimeout {
			get { return DefaultCommunicationTimeouts.Instance.OpenTimeout; }
		}

		public TimeSpan CloseTimeout {
			get { return close_timeout; }
			set { close_timeout = value; }
		}

		public TimeSpan OpenTimeout {
			get { return open_timeout; }
			set { open_timeout = value; }
		}

		public int ManualFlowControlLimit {
			get { return flow_limit; }
			set { flow_limit = value; }
		}

		public ServiceEndpoint AddServiceEndpoint (
			string implementedContract, Binding binding, string address)
		{
			return AddServiceEndpoint (implementedContract,
				binding,
				new Uri (address, UriKind.RelativeOrAbsolute));
		}

		[MonoTODO]
		public ServiceEndpoint AddServiceEndpoint (
			string implementedContract, Binding binding,
			string address, Uri listenUri)
		{
			Uri uri = new Uri (address, UriKind.RelativeOrAbsolute);
			return AddServiceEndpoint (
				implementedContract, binding, uri, uri);
		}

		[MonoTODO]
		public ServiceEndpoint AddServiceEndpoint (
			string implementedContract, Binding binding,
			Uri address)
		{
			return AddServiceEndpoint (implementedContract, binding, address, address);
		}

		[MonoTODO]
		public ServiceEndpoint AddServiceEndpoint (
			string implementedContract, Binding binding,
			Uri address, Uri listenUri)
		{
			EndpointAddress ea = BuildEndpointAddress (address, binding);
			ContractDescription cd = GetContract (implementedContract);
			if (cd == null)
				throw new InvalidOperationException (String.Format ("Contract '{0}' was not found in the implemented contracts in this service host.", implementedContract));
			return AddServiceEndpointCore (cd, binding, ea, listenUri);
		}

		Type PopulateType (string typeName)
		{
			Type type = Type.GetType (typeName);
			if (type != null)
				return type;
			foreach (ContractDescription cd in ImplementedContracts.Values) {
				type = cd.ContractType.Assembly.GetType (typeName);
				if (type != null)
					return type;
			}
			return null;
		}

		ContractDescription GetContract (string typeName)
		{
			//FIXME: hack hack hack
			ImplementedContracts ["IHttpGetHelpPageAndMetadataContract"] =
				ContractDescription.GetContract (typeof (IHttpGetHelpPageAndMetadataContract));

			// FIXME: As long as I tried, *only* IMetadataExchange
			// is the exception case that does not require full
			// type name. Hence I treat it as a special case.
			if (typeName == ServiceMetadataBehavior.MexContractName) {
				if (!Description.Behaviors.Contains (typeof (ServiceMetadataBehavior)) && Array.IndexOf (Description.ServiceType.GetInterfaces (), typeof (IMetadataExchange)) < 0)
					throw new InvalidOperationException (
						"Add ServiceMetadataBehavior to the ServiceHost to add a endpoint for IMetadataExchange contract.");
					
				ImplementedContracts [ServiceMetadataBehavior.MexContractName] =
					ContractDescription.GetContract (typeof (IMetadataExchange));

				foreach (ContractDescription cd in ImplementedContracts.Values)
					if (cd.ContractType == typeof (IMetadataExchange))
						return cd;
				return null;
			}

			Type type = PopulateType (typeName);
			if (type == null)
				return null;

			foreach (ContractDescription cd in ImplementedContracts.Values) {
				// FIXME: This check is a negative side effect 
				// of the above hack.
				if (cd.ContractType == typeof (IMetadataExchange))
					continue;

				if (cd.ContractType == type ||
				    cd.ContractType.IsSubclassOf (type) ||
				    type.IsInterface && cd.ContractType.GetInterface (type.FullName) == type)
					return cd;
			}
			return null;
		}

		internal EndpointAddress BuildEndpointAddress (Uri address, Binding binding)
		{
			if (!address.IsAbsoluteUri) {
				// Find a Base address with matching scheme,
				// and build new absolute address
				if (!base_addresses.Contains (binding.Scheme))
					throw new InvalidOperationException (String.Format ("Could not find base address that matches Scheme {0} for endpoint {1}", binding.Scheme, binding.Name));

				Uri baseaddr = base_addresses [binding.Scheme];

				if (!baseaddr.AbsoluteUri.EndsWith ("/"))
					baseaddr = new Uri (baseaddr.AbsoluteUri + "/");
				address = new Uri (baseaddr, address);
			}
			return new EndpointAddress (address);
		}

		internal ServiceEndpoint AddServiceEndpointCore (
			ContractDescription cd, Binding binding, EndpointAddress address, Uri listenUri)
		{
			foreach (ServiceEndpoint e in Description.Endpoints)
				if (e.Contract == cd)
					return e;
			ServiceEndpoint se = new ServiceEndpoint (cd, binding, address);
			se.ListenUri = listenUri.IsAbsoluteUri ? listenUri : new Uri (address.Uri, listenUri);
			Description.Endpoints.Add (se);
			return se;
		}

		[MonoTODO]
		protected virtual void ApplyConfiguration ()
		{
			if (Description == null)
				throw new InvalidOperationException ("ApplyConfiguration requires that the Description property be initialized. Either provide a valid ServiceDescription in the CreateDescription method or override the ApplyConfiguration method to provide an alternative implementation");

			ServiceElement service = GetServiceElement ();

			//TODO: Should we call here LoadServiceElement ?
			if (service != null) {
				
				//base addresses
				HostElement host = service.Host;
				foreach (BaseAddressElement baseAddress in host.BaseAddresses) {
					this.base_addresses.Add (new Uri (baseAddress.BaseAddress));
				}

				// services
				foreach (ServiceEndpointElement endpoint in service.Endpoints) {
					// FIXME: consider BindingName as well
					ServiceEndpoint se = AddServiceEndpoint (
						endpoint.Contract,
						ConfigUtil.CreateBinding (endpoint.Binding, endpoint.BindingConfiguration),
						endpoint.Address.ToString ());
				}
				// behaviors
				// TODO: use EvaluationContext of ServiceElement.
				ServiceBehaviorElement behavior = ConfigUtil.BehaviorsSection.ServiceBehaviors.Find (service.BehaviorConfiguration);
				if (behavior != null) {
					for (int i = 0; i < behavior.Count; i++) {
						BehaviorExtensionElement bxel = behavior [i];
						IServiceBehavior b = (IServiceBehavior) behavior [i].CreateBehavior ();
						if (b != null)
							Description.Behaviors.Add (b);
					}
				}
			}
			// TODO: consider commonBehaviors here

			// ensure ServiceAuthorizationBehavior
			Authorization = Description.Behaviors.Find<ServiceAuthorizationBehavior> ();
			if (Authorization == null) {
				Authorization = new ServiceAuthorizationBehavior ();
				Description.Behaviors.Add (Authorization);
			}

			// ensure ServiceDebugBehavior
			ServiceDebugBehavior debugBehavior = Description.Behaviors.Find<ServiceDebugBehavior> ();
			if (debugBehavior == null) {
				debugBehavior = new ServiceDebugBehavior ();
				Description.Behaviors.Add (debugBehavior);
			}
		}

		private ServiceElement GetServiceElement() {
			Type serviceType = Description.ServiceType;
			if (serviceType == null)
				return null;

			return ConfigUtil.ServicesSection.Services [serviceType.FullName];			
		}

		internal ContractDescription GetContract (string name, string ns)
		{
			foreach (ContractDescription d in ImplementedContracts.Values)
				if (d.Name == name && d.Namespace == ns)
					return d;
			return null;
		}

		protected abstract ServiceDescription CreateDescription (
			out IDictionary<string,ContractDescription> implementedContracts);

		[MonoTODO]
		protected void InitializeDescription (UriSchemeKeyedCollection baseAddresses)
		{
			this.base_addresses = baseAddresses;
			IDictionary<string,ContractDescription> retContracts;
			description = CreateDescription (out retContracts);
			contracts = retContracts;

			ApplyConfiguration ();
		}

		[MonoTODO]
		protected virtual void InitializeRuntime ()
		{
			//First validate the description, which should call all behaviors
			//'Validate' method.
			ValidateDescription ();
			
			//Build all ChannelDispatchers, one dispatcher per user configured EndPoint.
			//We must keep thet ServiceEndpoints as a seperate collection, since the user
			//can change the collection in the description during the behaviors events.
			Dictionary<ServiceEndpoint, ChannelDispatcher> endPointToDispatcher = new Dictionary<ServiceEndpoint,ChannelDispatcher>();
			ServiceEndpoint[] endPoints = new ServiceEndpoint[Description.Endpoints.Count];
			Description.Endpoints.CopyTo (endPoints, 0);
			foreach (ServiceEndpoint se in endPoints) {
				//Let all behaviors add their binding parameters
				BindingParameterCollection commonParams =
					new BindingParameterCollection ();
				AddBindingParameters (commonParams, se);
				ChannelDispatcher channel = BuildChannelDispatcher (se, commonParams);
				ChannelDispatchers.Add (channel);				
				endPointToDispatcher[se] = channel;				
			}

			//After the ChannelDispatchers are created, and attached to the service host
			//Apply dispatching behaviors.
			foreach (IServiceBehavior b in Description.Behaviors)
				b.ApplyDispatchBehavior (Description, this);

			foreach(KeyValuePair<ServiceEndpoint, ChannelDispatcher> val in endPointToDispatcher)
				ApplyDispatchBehavior(val.Value, val.Key);			
		}

		private void ValidateDescription () {
			foreach (IServiceBehavior b in Description.Behaviors)
				b.Validate (Description, this);
			foreach (ServiceEndpoint endPoint in Description.Endpoints)
				endPoint.Validate ();
		}		

		private void ApplyDispatchBehavior (ChannelDispatcher dispatcher, ServiceEndpoint endPoint) {			
			foreach (IContractBehavior b in endPoint.Contract.Behaviors)
				b.ApplyDispatchBehavior (endPoint.Contract, endPoint, dispatcher.Endpoints[0].DispatchRuntime);
			foreach (IEndpointBehavior b in endPoint.Behaviors)
				b.ApplyDispatchBehavior (endPoint, dispatcher.Endpoints [0]);
			foreach (OperationDescription operation in endPoint.Contract.Operations) {
				foreach (IOperationBehavior b in operation.Behaviors)
					b.ApplyDispatchBehavior (operation, dispatcher.Endpoints [0].DispatchRuntime.Operations [operation.Name]);
			}

		}

		internal ChannelDispatcher BuildChannelDispatcher (ServiceEndpoint se, BindingParameterCollection commonParams) {

			//User the binding parameters to build the channel listener and Dispatcher
			IChannelListener lf = BuildListener (se, commonParams);
			ChannelDispatcher cd = new ChannelDispatcher (
				lf, se.Binding.Name);
			cd.MessageVersion = se.Binding.MessageVersion;
			if (cd.MessageVersion == null)
				cd.MessageVersion = MessageVersion.Default;

			//Attach one EndpointDispacher to the ChannelDispatcher
			EndpointDispatcher endpoint_dispatcher =
				new EndpointDispatcher (se.Address, se.Contract.Name, se.Contract.Namespace);
			endpoint_dispatcher.ContractFilter = GetContractFilter (se.Contract);
			endpoint_dispatcher.ChannelDispatcher = cd;
			cd.Endpoints.Add (endpoint_dispatcher);
			
			//Build the dispatch operations
			DispatchRuntime db = endpoint_dispatcher.DispatchRuntime;
			foreach (OperationDescription od in se.Contract.Operations)
				if (!db.Operations.Contains (od.Name))
					PopulateDispatchOperation (db, od);

			return cd;
		}

		private MessageFilter GetContractFilter (ContractDescription contractDescription)
		{
			List<string> actions = new List<string> ();
			foreach (OperationDescription od in contractDescription.Operations)
				foreach (MessageDescription md in od.Messages)
					if (md.Direction == MessageDirection.Input)
						if (md.Action == "*")
							return new MatchAllMessageFilter ();
						else
							actions.Add (md.Action);

			return new ActionMessageFilter (actions.ToArray ());
		}

		private void AddBindingParameters (BindingParameterCollection commonParams, ServiceEndpoint endPoint) {

			commonParams.Add (ChannelProtectionRequirements.CreateFromContract (endPoint.Contract));
			foreach (IServiceBehavior b in Description.Behaviors)
				b.AddBindingParameters (Description, this, Description.Endpoints, commonParams);

			foreach (IContractBehavior b in endPoint.Contract.Behaviors)
				b.AddBindingParameters (endPoint.Contract, endPoint, commonParams);
			foreach (IEndpointBehavior b in endPoint.Behaviors)
				b.AddBindingParameters (endPoint, commonParams);
			foreach (OperationDescription operation in endPoint.Contract.Operations) {
				foreach (IOperationBehavior b in operation.Behaviors)
					b.AddBindingParameters (operation, commonParams);
			}
		}

		void PopulateDispatchOperation (DispatchRuntime db, OperationDescription od) {
			string reqA = null, resA = null;
			foreach (MessageDescription m in od.Messages) {
				if (m.Direction == MessageDirection.Input)
					reqA = m.Action;
				else
					resA = m.Action;
			}
			DispatchOperation o =
				od.IsOneWay ?
				new DispatchOperation (db, od.Name, reqA) :
				new DispatchOperation (db, od.Name, reqA, resA);
			bool has_void_reply = false;
			foreach (MessageDescription md in od.Messages) {
				if (md.Direction == MessageDirection.Input &&
					md.Body.Parts.Count == 1 &&
					md.Body.Parts [0].Type == typeof (Message))
					o.DeserializeRequest = false;
				if (md.Direction == MessageDirection.Output &&
					md.Body.ReturnValue != null) {
					if (md.Body.ReturnValue.Type == typeof (Message))
						o.SerializeReply = false;
					else if (md.Body.ReturnValue.Type == typeof (void))
						has_void_reply = true;
				}
			}

			// Setup Invoker
			// FIXME: support async method
			if (od.SyncMethod != null)
				o.Invoker = new SyncMethodInvoker (od.SyncMethod);
			else
				o.Invoker = new AsyncMethodInvoker (od.BeginMethod, od.EndMethod);

			// Setup Formater
			o.Formatter = BaseMessagesFormatter.Create (od);

			if (o.Action == "*" && o.ReplyAction == "*") {
				//Signature : Message  (Message)
				//	    : void  (Message)
				//FIXME: void (IChannel)
				if (!o.DeserializeRequest && (!o.SerializeReply || has_void_reply))
					db.UnhandledDispatchOperation = o;
			}

			db.Operations.Add (o);
		}

		[MonoTODO]
		protected void LoadConfigurationSection (ServiceElement element)
		{
			ServicesSection services = ConfigUtil.ServicesSection;
		}

		void DoOpen (TimeSpan timeout)
		{
			for (int i = 0; i < ChannelDispatchers.Count; i++)
				ChannelDispatchers [i].Open (timeout);
		}

		IChannelListener BuildListener (ServiceEndpoint se,
			BindingParameterCollection pl)
		{
			Binding b = se.Binding;
			if (b.CanBuildChannelListener<IReplySessionChannel> (pl))
				return b.BuildChannelListener<IReplySessionChannel> (se.ListenUri, "", se.ListenUriMode, pl);
			if (b.CanBuildChannelListener<IReplyChannel> (pl))
				return b.BuildChannelListener<IReplyChannel> (se.ListenUri, "", se.ListenUriMode, pl);
			if (b.CanBuildChannelListener<IInputSessionChannel> (pl))
				return b.BuildChannelListener<IInputSessionChannel> (se.ListenUri, "", se.ListenUriMode, pl);
			if (b.CanBuildChannelListener<IInputChannel> (pl))
				return b.BuildChannelListener<IInputChannel> (se.ListenUri, "", se.ListenUriMode, pl);
			throw new InvalidOperationException ("None of the listener channel types is supported");
		}

		[MonoTODO]
		protected override sealed void OnAbort ()
		{
		}

		[MonoTODO]
		protected override sealed IAsyncResult OnBeginClose (
			TimeSpan timeout, AsyncCallback callback, object state)
		{
			throw new NotImplementedException ();
		}

		[MonoTODO]
		protected override sealed IAsyncResult OnBeginOpen (
			TimeSpan timeout, AsyncCallback callback, object state)
		{
			throw new NotImplementedException ();
		}

		[MonoTODO]
		protected override void OnClose (TimeSpan timeout)
		{
			ReleasePerformanceCounters ();
			List<ChannelDispatcherBase> l = new List<ChannelDispatcherBase> (ChannelDispatchers);
			foreach (ChannelDispatcherBase e in l)
				e.Close ();
		}

		protected override sealed void OnOpen (TimeSpan timeout)
		{
			InitializeRuntime ();
			DoOpen (timeout);
		}

		[MonoTODO]
		protected override void OnEndClose (IAsyncResult result)
		{
			throw new NotImplementedException ();
		}

		[MonoTODO]
		protected override sealed void OnEndOpen (IAsyncResult result)
		{
			throw new NotImplementedException ();
		}

		protected override void OnOpened ()
		{
		}

		[MonoTODO]
		protected void ReleasePerformanceCounters ()
		{
		}

		void IDisposable.Dispose ()
		{
			Close ();
		}

		class SyncMethodInvoker : IOperationInvoker
		{
			readonly MethodInfo _methodInfo;
			public SyncMethodInvoker (MethodInfo methodInfo) {
				_methodInfo = methodInfo;
			}
			
			#region IOperationInvoker Members

			public bool IsSynchronous {
				get { return true; }
			}

			public object [] AllocateParameters () {
				return new object [_methodInfo.GetParameters ().Length];
			}

			public object Invoke (object instance, object [] parameters)
            {
				return _methodInfo.Invoke (instance, parameters);
			}

			public IAsyncResult InvokeBegin (object instance, object [] inputs, AsyncCallback callback, object state) {
				throw new NotSupportedException ();
			}

			public object InvokeEnd (object instance, out object [] outputs, IAsyncResult result) {
				throw new NotSupportedException ();
			}

			#endregion
		}

		class AsyncMethodInvoker : IOperationInvoker
		{
			readonly MethodInfo _beginMethodInfo, _endMethodInfo;
			public AsyncMethodInvoker (MethodInfo beginMethodInfo, MethodInfo endMethodInfo) {
				_beginMethodInfo = beginMethodInfo;
				_endMethodInfo = endMethodInfo;
			}

			#region IOperationInvoker Members

			public bool IsSynchronous {
				get { return false; }
			}

			public object [] AllocateParameters () {
				return new object [_beginMethodInfo.GetParameters ().Length - 2 + _endMethodInfo.GetParameters().Length-1];
			}

			public object Invoke (object instance, object [] parameters) {
				throw new NotImplementedException ("Can't invoke async method synchronously");
				//BUGBUG: need to differentiate between input and output parameters.
				IAsyncResult asyncResult = InvokeBegin(instance, parameters, delegate(IAsyncResult ignore) { }, null);
				asyncResult.AsyncWaitHandle.WaitOne();
				return InvokeEnd(instance, out parameters, asyncResult);
			}

			public IAsyncResult InvokeBegin (object instance, object [] inputs, AsyncCallback callback, object state) {
				if (inputs.Length + 2 != _beginMethodInfo.GetParameters ().Length)
					throw new ArgumentException ("Wrong number of input parameters");
				object [] fullargs = new object [_beginMethodInfo.GetParameters ().Length];
				Array.Copy (inputs, fullargs, inputs.Length);
				fullargs [inputs.Length] = callback;
				fullargs [inputs.Length + 1] = state;
				return (IAsyncResult) _beginMethodInfo.Invoke (instance, fullargs);
			}

			public object InvokeEnd (object instance, out object [] outputs, IAsyncResult asyncResult) {
				outputs = new object [_endMethodInfo.GetParameters ().Length - 1];
				object [] fullargs = new object [_endMethodInfo.GetParameters ().Length];
				fullargs [outputs.Length] = asyncResult;
				object result = _endMethodInfo.Invoke (instance, fullargs);
				Array.Copy (fullargs, outputs, outputs.Length);
				return result;
			}

			#endregion
		}
	}
}
